-- SRAM interface for XSV
-- every communication between FPGA and external SRAM should use this interface
-- see sram.doc

library IEEE;
use IEEE.std_logic_1164.all; 

--library VIRTEX;
--use VIRTEX.all; -- for xilinx component signals


entity sram is
    generic (adr_width: integer:= 19; -- width of the address bus
    	data_width: integer:= 16  -- width of the data bus
    	); 
    port ( -- user signals
	    clk: in STD_LOGIC; -- clock signal - every user signal is latched rising edge of clk
		  arst: in std_logic; -- global asynchronous set reset 
      wr_rdn: in STD_LOGIC; -- write / not read
		  adri: in STD_LOGIC_VECTOR (0 to adr_width-1); -- address from which data will be read or written
		  datar: out STD_LOGIC_VECTOR (0 to data_width-1); -- data read form the sram
      dataw: in STD_LOGIC_VECTOR (0 to data_width-1); -- data to be written to the sram
		  ce_adr: in std_logic; -- clock enable for address flip-flops (except the LSB which always always '1')
		  ce_datar: in std_logic; -- clock enable for datar flip-flops
		  ce_dataw: in std_logic_vector(0 to data_width/8 -1); -- clock enable for write flip-flops - seperate for each byte
      -- external SRAM signals
      adr: out STD_LOGIC_VECTOR (0 to adr_width-1); -- address bus 
		  data_i: in STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (input)
		  data_o: out STD_LOGIC_VECTOR (0 to data_width-1); -- data bus (output)
		  data_t: out STD_LOGIC; -- data bus (tri-state)
      oen: out std_logic; -- output enable (active low)
      wen: out std_logic; -- write enable
      cen: out std_logic; -- chip enable (always zero)
      ubn, lbn: out std_logic -- byte enable lower and uppper (active low)
    );
end sram;

architecture sram_arch of sram is
  signal datawq: std_logic_vector(data_width-1 downto 0); -- dataw in the next clk cycle
  signal wrq, wrq_clr: std_logic; -- signals that control wrpn and oe_data
  signal wrqq, wrqq_clr: std_logic; -- additional signal for oe_data and wenp
  signal data_hiZ: std_logic; -- HiZ control for datap
  signal oeq: std_logic; -- axuliary signal for oe_data
begin
  -- address flip-flops (in IOB) and multiplexer
process (clk, arst) begin
	if arst= '1' then
	  adr<= (others=>'0');	
    elsif clk'event and clk='1' then -- insert flip-flops (FFs) without reset (output FFs)
	  adr(adr_width-1)<= adri(adr_width-1); -- always transfer the LSB (disregarding the ce_adr)
	  if ce_adr='1' then 
	    adr(0 to adr_width-2)<= adri(0 to adr_width-2); -- select write address	
	  end if;
	end if;
  end process;

-- data path flip-flops insertet in IOB
process (clk, arst) begin
    if arst='1' then
		  data_o<= (others=> '0');
		  datar<= (others=>'0');
	elsif clk'event and clk='1' then -- insert flip-flops (FFs) without reset (output FFs)
	  if ce_datar='1' then
        datar<= data_i;
	  end if;
    for i in 0 to data_width/8 -1 loop
	    if ce_dataw(i)='1' then
		    data_o(8*i to 8*i +7)<= dataw(8*i to 8*i+7);
		  end if;
    end loop;
	end if;
  end process;
  
  -- control signal logic
  cen<= '0'; -- SRAM is always selected
  
  -- OEPN flip-flop
  process (clk, arst) begin
	if arst='1' then
		oen<= '0';
    ubn<= '1'; lbn<= '1';
    elsif clk'event and clk='1' then
      oen<= wr_rdn; -- set SRAM in HiZ whenever write clock cycle
      if wr_rdn='1' then -- write operation
        lbn<= not ce_dataw(0);
      else -- read operation
        lbn<= '0';
      end if;
      if data_width>8 then -- two byte operation
        if wr_rdn='1' then -- write operation
          ubn<= not ce_dataw(1);
        else -- read operation
          ubn<= '0';
        end if;
      end if;
	end if;
  end process;

  -- wrq flip-flop with asynch. reset
  wen<= not wrq;
  
  process (wrq_clr, arst, clk) begin
     if wrq_clr= '1' or arst='1' then -- asynch. reset, this reset causes additional delay of wrq signal after clk goes low
       wrq<= '0';
     elsif clk'event and clk='1' then -- synch. logic
       wrq<= wr_rdn;
     end if;
  end process;
  
  -- wrqq_clr logic - clears wrq when clk low  
  wrq_clr<= wrqq and wrq; -- the reset signal that is activeted when clock is low and until wrq is high
  --wrqq_clr<= wrqq and not wrq; -- signal wrqq_clr has an additional delay in comparison to wr_clr
  
  process (clk, arst, wrq_clr) begin
    if arst='1' or wrq_clr= '1' then
		wrqq<= '0';
	elsif clk'event and clk='0' then -- falling edge of clk!
		wrqq<= wrq;
	end if;
  end process;

  data_HiZ<= not (wrqq or wrq); -- data_hiz might be the same signal as wrq however this additional logic 
  -- causes additional delay (datap must be in LowZ when WEPN is low)
  data_t<= data_HiZ;
  
    
end sram_arch;
